<?php

namespace Drupal\config_enforce_devel;

use Drupal\Core\File\FileSystemInterface;
use Drupal\config_enforce\ConfigEnforce;

/**
 * Enforces configuration.
 */
class ConfigEnforceDevel extends ConfigEnforce {

  /* The name of a config object to enforce. */
  protected $configName;

  /* The enforcement level for a config object. */
  protected $enforcementLevel;

  /* The module in which to write a config file. */
  protected $targetModule;

  /* The directory in which to write a config file. */
  protected $configDirectory;

  /* The path to write a config file. */
  protected $configFilePath;

  /* The URI of the host form. */
  protected $configFormUri = '';

  /**
   * Provide Config Devel module with configs to read from disk.
   *
   * @see hook_config_devel_auto_import()
   */
  public static function autoImport() {
    $auto_import = [];
    foreach (self::getEnforcedConfigs() as $config_name => $setting) {
      $auto_import[$config_name] = [
        'filename' => $setting['config_file_path'],
        'hash' => isset($setting['hash']) ? $setting['hash'] : '',
      ];
    }
    return $auto_import;
  }

  /**
   * Provide Config Devel module with config files to write to disk.
   *
   * @see hook_config_devel_auto_export()
   */
  public static function autoExport() {
    $auto_export = [];
    foreach (self::getEnforcedConfigs() as $config_name => $settings) {
      $auto_export[] = $settings['config_file_path'];
    }
    return $auto_export;
  }

  /**
   * Save hashes of auto-imported config files.
   *
   * @see config_enforce_devel_config_devel_post_auto_import()
   */
  public static function postAutoImport(array $auto_import) {
    $enforced_configs = self::getEditableEnforcedConfigs();
    foreach ($auto_import as $config_name => $info) {
      $config = $enforced_configs->get(self::encode($config_name));
      $config['hash'] = $info['hash'];
      $enforced_configs->set(self::encode($config_name), $config);
    }
    $enforced_configs->save();
  }

  /**
   * Set the config object name attribute.
   */
  public function setConfigName(string $config_name) {
    $this->configName = $config_name;
    return $this;
  }

  /**
   * Set the enforcement level attribute.
   */
  public function setEnforcementLevel(string $level) {
    $this->enforcementLevel = $level;
    return $this;
  }

  /**
   * Set the target module attribute.
   */
  public function setTargetModule(string $module) {
    $this->targetModule = $module;
    return $this;
  }

  /**
   * Set the config directory attribute.
   */
  public function setConfigDirectory(string $directory) {
    $this->configDirectory = $directory;
    return $this;
  }

  /**
   * Set the config directory attribute.
   */
  public function setConfigFormUri(string $uri) {
    $this->configFormUri = $uri;
    return $this;
  }

  /**
   * Remove a config object from the enforced configs.
   */
  public function clear() {
    self::getEditableEnforcedConfigs()
      ->clear(self::encode($this->configName))
      ->save();
    return $this;
  }

  /**
   * Add or update a config object in the enforced configs.
   */
  public function save() {
    $this
      ->setConfigFilePath()
      ->prepareConfigDirectory()
      ->initializeNewConfigFile()
      ->cleanStaleConfigFile();

    self::getEditableEnforcedConfigs()
      ->set(self::encode($this->configName), [
        'target_module' => $this->targetModule,
        'config_directory' => $this->configDirectory,
        'config_file_path' => $this->configFilePath,
        'enforcement_level' => $this->enforcementLevel,
        'config_form_uri' => $this->configFormUri,
      ])
      ->save();
    return $this;
  }

  /**
   * Return this module's stored config.
   */
  protected static function getEditableEnforcedConfigs() {
    return \Drupal::configFactory()
      ->getEditable(self::CONFIG_NAME);
  }

  /**
   * Delete an existing config file, if its path has changed.
   */
  protected function cleanStaleConfigFile() {
    if ($this->configFilePathChanged()) {
      \Drupal::service('file_system')
        ->delete(self::getConfigFilePath($this->configName));
    }
    return $this;
  }

  /**
   * Create a new config file.
   *
   * Config Devel is generally responsible for writing enforced config to disk.
   * However, when we enable a config for the first time, it won't yet be in
   * the list of enforced configs the Config Devel should write to disk. So, we
   * have to write it ourselves this time.
   *
   * @see ConfigEnforceDevel::autoExport()
   */
  protected function initializeNewConfigFile() {
    if ($this->configFilePathChanged()) {
      \Drupal::service('config_devel.writeback_subscriber')
        ->writeBackConfig(\Drupal::config($this->configName), [$this->configFilePath]);
    }
    return $this;
  }

  /**
   * Determine whether the path for a config file has changed (or is new).
   */
  protected function configFilePathChanged() {
    $old_path = ConfigEnforce::getConfigFilePath($this->configName);
    if (!file_exists($old_path)) return TRUE;
    return $this->configFilePath !== $old_path;
  }

  /**
   * Ensure that the directory, in which a config file should be written, exists.
   */
  protected function prepareConfigDirectory() {
    $file_system = \Drupal::service('file_system');
    $directory = $file_system->dirname($this->configFilePath);
    $options = FileSystemInterface::CREATE_DIRECTORY + FileSystemInterface::MODIFY_PERMISSIONS;
    $file_system->prepareDirectory($directory, $options);
    return $this;
  }

  /**
   * Derive the config file path from the target module and config directory.
   */
  protected function setConfigFilePath() {
    $this->configFilePath = drupal_get_path('module', $this->targetModule);
    $this->configFilePath .= '/' . $this->configDirectory . '/';
    $this->configFilePath .= $this->configName . '.yml';
    return $this;
  }

}
