<?php

namespace Drupal\config_enforce_devel\Form;

use Drupal\config_enforce\ConfigEnforce;
use Drupal\config_enforce\Form\FormHelperTrait;
use Drupal\config_enforce_devel\ConfigEnforceDevel;
use Drupal\Core\Config\FileStorage;
use Drupal\Core\Config\InstallStorage;
use Drupal\Core\Config\StorageInterface;
use Drupal\Core\Form\FormStateInterface;

/**
 * Defines a settings form for the Config Enforce Devel module.
 */
trait DevelFormHelperTrait {

  use FormHelperTrait;

  // The current config object name.
  protected $current_config;

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $this->setSharedFormProperties($form, $form_state);

    $this->saveConfig();

    return parent::submitForm($this->form(), $this->formState());
  }

  /**
   * Add fields for a single config object.
   */
  protected function addEnforcedConfigItem($config) {
    parent::addEnforcedConfigItem($config);

    $this
      ->setCurrentConfig($config)
      ->addConfigEnforceEnabledField()
      ->overrideTargetModuleField()
      ->overrideConfigDirectoryField()
      ->overrideEnforcementLevelField()
      ->addEnforceDependenciesFields()
      ->setConfigEnforcePathStates();
  }

  /**
   * Add a checkbox to control whether the current config object shouldbe enforced.
   */
  protected function addConfigEnforceEnabledField() {
    $this->form()['config_enforce'][$this->getCurrentConfig()]['config_enforce_enabled'] = [
      '#type' => 'checkbox',
      '#title' => 'Enforce config',
      '#description' => "Enable enforcement of the config from this form.",
      '#default_value' => $this->currentConfigIsEnforced(),
      '#attributes' => [
        'config-enforce-enabled' => $this->getEnabledAttribute(),
      ],
    ];
    return $this;
  }

  /**
   * Add a field to allow the target module to be selected.
   */
  protected function overrideTargetModuleField() {
    $config = $this->getCurrentConfig();
    $this->form()['config_enforce'][$config]['target_module'] = $this->getTargetModuleField();
    $this->form()['config_enforce'][$config]['target_module']['#states'] = $this->getEnabledVisibilityState();
    if (ConfigEnforce::isEnforced($config)) {
      $module = ConfigEnforce::getTargetModule($config);
      $this->form()['config_enforce'][$config]['target_module']['#title'] .= ' (<em>' . $module . '</em>)';
      $this->form()['config_enforce'][$config]['target_module']['#default_value'] = $module;
    }
    return $this;
  }

  /**
   * Add a field to allow the config directory to be selected.
   */
  protected function overrideConfigDirectoryField() {
    $config = $this->getCurrentConfig();
    $this->form()['config_enforce'][$config]['config_directory'] = $this->getConfigDirectoryField();
    $this->form()['config_enforce'][$config]['config_directory']['#states'] = $this->getEnabledVisibilityState();
    if (ConfigEnforce::isEnforced($config)) {
      $directory = ConfigEnforce::getConfigDirectory($config);
      $this->form()['config_enforce'][$config]['config_directory']['#title'] .= ' (<em>' . $directory . '</em>)';
      $this->form()['config_enforce'][$config]['config_directory']['#default_value'] = $directory;
    }
    return $this;
  }

  /**
   * Add a field to allow the enforcement level directory to be specified.
   */
  protected function overrideEnforcementLevelField() {
    $config = $this->getCurrentConfig();
    $this->form()['config_enforce'][$config]['enforcement_level'] = $this->getEnforcementLevelField();
    $this->form()['config_enforce'][$config]['enforcement_level']['#states'] = $this->getEnabledVisibilityState();
    if ($this->currentConfigIsEnforced()) {
      $this->form()['config_enforce'][$config]['enforcement_level']['#title'] .= ' (<em>' . ConfigEnforce::getEnforcementLevelLabel($config) . '</em>)';
      $this->form()['config_enforce'][$config]['enforcement_level']['#default_value'] = ConfigEnforce::getEnforcementLevel($config);
    }
    return $this;
  }

  /**
   * Add a checkbox to control whether to enforce dependent config objects.
   */
  protected function addEnforceDependenciesFields() {
    $dependencies = $this->getConfigDependencies();
    if (!count($dependencies) > 0) return $this;

    $this->form()['config_enforce'][$this->getCurrentConfig()]['enforce_dependencies'] = $this->getEnforceDependenciesField();
    $this->form()['config_enforce'][$this->getCurrentConfig()]['enforce_dependencies']['#attributes'] = [
      'config-enforce-dependencies' => $this->getEnforceDependenciesAttribute(),
    ];
    $this->form()['config_enforce'][$this->getCurrentConfig()]['enforce_dependencies']['#states'] = $this->getEnabledVisibilityState();

    $this->form()['config_enforce'][$this->getCurrentConfig()]['dependencies'] = [
      '#type' => 'container',
      '#states' => $this->getEnforceDependenciesVisibilityState(),
    ];
    $enforced_dependencies = 0;
    foreach ($dependencies as $dependency) {
      $enforced = ConfigEnforce::isEnforced($dependency);
      if ($enforced) $enforced_dependencies++;

      $this->form()['config_enforce'][$this->getCurrentConfig()]['dependencies'][$dependency] = [
        '#type' => 'checkbox',
        '#title' => $dependency . ($enforced ? $this->t(' (already enforced)') : ''),
        '#default_value' => FALSE,
        '#states' => [
          'checked' => [
            ':input[config-enforce-dependencies="' . $this->getEnforceDependenciesAttribute() . '"]' => ['checked' => TRUE],
          ],
        ],
      ];
    }
    if ($enforced_dependencies > 0) {
      $this->disableEnforcedDependencies();
    }
    return $this;
  }

  /**
   * Disable checkboxes for dependencies that are already enforced, but allow for them to be overridden.
   */
  protected function disableEnforcedDependencies() {
    $this->form()['config_enforce'][$this->getCurrentConfig()]['override_existing_dependencies'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Override existing enforcement settings'),
      '#description' => $this->t('Allow existing enforcement settings for dependencies to be overridden by the settings from this form.'),
      '#default_value' => FALSE,
      '#attributes' => [
        'config-enforce-override-dependencies' => $this->getEnforceDependenciesAttribute(),
      ],
      '#states' => $this->getEnforceDependenciesVisibilityState(),
    ];

    foreach ($this->getConfigDependencies() as $dependency) {
      if (!ConfigEnforce::isEnforced($dependency)) continue;
      $this->form()['config_enforce'][$this->getCurrentConfig()]['dependencies'][$dependency]['#states'] = [
        'enabled' => [
          ':input[config-enforce-override-dependencies="' . $this->getEnforceDependenciesAttribute() . '"]' => ['checked' => TRUE],
        ],
      ];
    }
  }

  /**
   * Set the current config object name.
   */
  protected function setCurrentConfig($config) {
    $this->current_config = $config;
    return $this;
  }

  /**
   * Return the current config object name.
   */
  protected function getCurrentConfig() {
    return $this->current_config;
  }

  /**
   * Set the Form API '#states' setting for the config file path.
   */
  protected function setConfigEnforcePathStates() {
    if ($this->currentConfigIsEnforced()) {
      $this->form()['config_enforce'][$this->getCurrentConfig()]['config_enforce_path']['#states'] = $this->getEnabledVisibilityState();
    }
    return $this;
  }

  /**
   * Determine whether the current config object is being enforced.
   */
  protected function currentConfigIsEnforced() {
    return ConfigEnforce::isEnforced($this->getCurrentConfig());
  }

  /**
   * Return the config-specific "enabled" attribute value.
   *
   * @see addConfigEnforceEnabledField()
   * @see getEnabledVisibilityState()
   */
  protected function getEnabledAttribute() {
    return 'config_enforce_enabled-' . $this->getCurrentConfig();
  }

  /**
   * Return the Form API '#states' setting for config fields.
   *
   * This makes a given config's fields' visibility depend on the 'enabled' field for that config.
   */
  protected function getEnabledVisibilityState() {
    return [
      'invisible' => [
        ':input[config-enforce-enabled="' . $this->getEnabledAttribute() . '"]' => ['checked' => FALSE],
      ],
    ];
  }

  /**
   * Return the config-specific "enabled" attribute value.
   *
   * @see foo()
   * @see getEnforceDependenciesVisibilityState()
   */
  protected function getEnforceDependenciesAttribute() {
    return 'config_enforce_dependencies-' . $this->getCurrentConfig();
  }

  /**
   * Return the Form API '#states' setting for config fields.
   *
   * This makes a given config's fields' visibility depend on the 'enabled' field for that config.
   */
  protected function getEnforceDependenciesVisibilityState() {
    return [
      'visible' => [
        ':input[config-enforce-dependencies="' . $this->getEnforceDependenciesAttribute() . '"]' => ['checked' => TRUE],
        ':input[config-enforce-enabled="' . $this->getEnabledAttribute() . '"]' => ['checked' => TRUE],
      ],
    ];
  }

  /**
   * Save our config.
   */
  protected function saveConfig() {
    $this->configEnforce = new ConfigEnforceDevel();
    foreach ($this->getFormConfigs() as $config) {
      $this->configEnforce->setConfigName($config);

      $this->setCurrentConfig($config);

      if (!$this->getEnforced()) {
        $this->configEnforce->clear();
        continue;
      }

      $this->prepareConfig();

      $this->configEnforce->save();

      if ($this->getEnforceDependencies()) {
        $this->saveConfigDependencies();
      }
    }
  }

  /**
   * Save enforcement settings for dependencies.
   */
  protected function saveConfigDependencies() {
    foreach ($this->getConfigDependencies() as $dependency) {
      if (!$this->getDependencyEnforced($dependency)) continue;

      $this->configEnforce->setConfigName($dependency);

      $this->prepareConfig();

      // Unset the URI, since this is not the form where the dependency is configured.
      // @TODO Preserve URI, if it's already set.
      $this->configEnforce->setConfigFormUri('');

      $this->configEnforce->save();
    }
  }

  /**
   * Prepare our config.
   */
  protected function prepareConfig() {
    $this->configEnforce
      ->setTargetModule($this->getTargetModule())
      ->setConfigDirectory($this->getConfigDirectory())
      ->setEnforcementLevel($this->getEnforcementLevel());
  }

  /**
   * Return a list of modules.
   */
  protected function getAllModules() {
    return ConfigEnforce::getAllModules();
  }

  /**
   * Return a list of available modules to target for writing config.
   */
  protected function getAvailableModules() {
    $modules = [];
    $available_modules = $this->config('config_enforce_devel.settings')->get('available_modules');
    $available_modules = !is_null($available_modules) ? $available_modules : [];
    foreach ($available_modules as $module) {
      $modules[$module] = $this->getAllModules()[$module];
    }
    return $modules;
  }

  /**
   * Return a human-readable name for a given module.
   */
  protected function getModuleName($module) {
    return $this->getAvailableModules()[$module];
  }

  /**
   * Return a list of directories for writing config.
   */
  protected function getAvailableConfigDirectories() {
    // @TODO Look this up in the storage service or make this a hook.
    return [
      InstallStorage::CONFIG_INSTALL_DIRECTORY => 'Install',
      InstallStorage::CONFIG_OPTIONAL_DIRECTORY => 'Optional',
    ];
  }

  /**
   * Return a human-readable name for a given config directory.
   */
  protected function getConfigDirectoryLabel($directory) {
    return $this->getAvailableConfigDirectories()[$directory];
  }

  /**
   * Return a list of all config object names from given module and directory.
   */
  protected function getModuleConfigNames($module, $directory) {
    $module_config_path = drupal_get_path('module', $module) . '/' . $directory;
    $storage = new FileStorage($module_config_path, StorageInterface::DEFAULT_COLLECTION);
    return $storage->listAll();
  }

  /**
   * Return array of config enforce values from the form.
   */
  protected function getConfigEnforceValuesFromForm() {
    return $this->formState()->getValue('config_enforce');
  }

  /**
   * Return the value of a specific field for the current config object.
   */
  protected function getConfigEnforceValue(string $field) {
    $values = $this->getConfigEnforceValuesFromForm();
    if ($this->areConfigEnforceValuesWrapped()) {
      // Some forms appear to wrap our settings in an array.
      $values = $values['config_enforce'];
    }
    return $values[$this->getCurrentConfig()][$field];
  }

  /**
   * Determine whether config enforce values are wrapped in an extra array.
   */
  protected function areConfigEnforceValuesWrapped() {
    $values = $this->getConfigEnforceValuesFromForm();
    return array_key_exists('config_enforce', $values);
  }

  /**
   * Return the new enabled setting from the form state.
   */
  protected function getEnforced() {
    return $this->getConfigEnforceValue('config_enforce_enabled');
  }

  /**
   * Return the new target module from the form state.
   */
  protected function getTargetModule() {
    return $this->getConfigEnforceValue('target_module');
  }

  /**
   * Return the new config directory from the form state.
   */
  protected function getConfigDirectory() {
    return $this->getConfigEnforceValue('config_directory');
  }

  /**
   * Return the new enforcement level from the form state.
   */
  protected function getEnforcementLevel() {
    return $this->getConfigEnforceValue('enforcement_level');
  }

  /**
   * Return whether to enforce dependencies based on the form state.
   */
  protected function getEnforceDependencies() {
    return $this->getConfigEnforceValue('enforce_dependencies');
  }

  /**
   * Return whether to enforce a given dependency based on the form state.
   */
  protected function getDependencyEnforced($dependency) {
    return $this->getConfigEnforceValue('dependencies')[$dependency];
  }

  /**
   * Return the basic definition for a target module field.
   */
  protected function getTargetModuleField() {
    return [
      '#type' => 'select',
      '#title' => $this->t('Target module'),
      '#description' => "The module to which config should be written.",
      '#options' => $this->getAvailableModules(),
      '#default_value' => $this->config('config_enforce_devel.settings')->get('defaults')['target_module'],
    ];
  }

  /**
   * Return the basic definition for a config directory field.
   */
  protected function getConfigDirectoryField() {
    $saved_value = $this->config('config_enforce_devel.settings')->get('defaults')['config_directory'];
    $default = !is_null($saved_value) ? $saved_value : InstallStorage::CONFIG_OPTIONAL_DIRECTORY;
    return [
      '#type' => 'select',
      '#title' => $this->t('Directory'),
      '#description' => "The config directory to which config should be written.",
      '#options' => $this->getAvailableConfigDirectories(),
      '#default_value' => $default,
    ];
  }

  /**
   * Return the basic definition for an enforcement level field.
   */
  protected function getEnforcementLevelField() {
    $saved_value = $this->config('config_enforce_devel.settings')->get('defaults')['enforcement_level'];
    $default = !is_null($saved_value) ? $saved_value : ConfigEnforce::CONFIG_ENFORCE_READONLY;
    return [
      '#type' => 'select',
      '#title' => $this->t('Enforcement level'),
      '#description' => "The level of enforcement to apply to config objects.",
      '#options' => ConfigEnforce::getEnforcementLevels(),
      '#default_value' => $default,
    ];
  }

  /**
   * Add a checkbox to control whether to enforce dependent config objects.
   */
  protected function getEnforceDependenciesField() {
    $saved_value = $this->config('config_enforce_devel.settings')->get('defaults')['enforce_dependencies'];
    $default = !is_null($saved_value) ? $saved_value : TRUE;
    return [
      '#type' => 'checkbox',
      '#title' => 'Enforce dependencies',
      '#description' => "Generate enforcement settings for the config objects upon which this one depends.",
      '#default_value' => $default,
    ];
  }

  /**
   * List unenforced config object names that have not been flagged for ignoring.
   */
  protected function getUnenforcedConfigs() {
    $unenforced_configs = $this->getAllUnenforcedConfigs();
    $ignored_configs = $this->getIgnoredConfigs();
    $filtered_configs = array_diff($unenforced_configs, $ignored_configs);
    return array_combine($filtered_configs, $filtered_configs);
  }

  /**
   * List all unenforced config object names.
   */
  protected function getAllUnenforcedConfigs() {
    $active_storage_configs = $this->configFactory()->listAll();
    $enforced_configs = array_keys(ConfigEnforce::getEnforcedConfigs());
    $unenforced_configs = array_diff($active_storage_configs, $enforced_configs);
    return array_combine($unenforced_configs, $unenforced_configs);
  }

  /**
   * List ignored config objects.
   */
  protected function getIgnoredConfigs() {
    $ignored_configs = $this->config('config_enforce_devel.settings')->get('ignored_configs');
    return is_null($ignored_configs) ? [] : array_combine($ignored_configs, $ignored_configs);
  }

  /**
   * Return a list of all config objects that the current config object depends upon.
   */
  protected function getConfigDependencies($known_deps = []) {
    $config = $this->getCurrentConfig();
    $all_deps = $this->config($config)->get('dependencies');
    if (empty($all_deps)) return [];
    $config_deps = array_key_exists('config', $all_deps) ? $all_deps['config'] : [];
    // Compile a list of known dependencies, so that we can avoid infinite recursion below.
    $known_deps = array_merge($known_deps, $config_deps);
    foreach ($config_deps as $config_dep) {
      // Temporarily set the current config to the dependency.
      $this->setCurrentConfig($config_dep);
      // Recurse to ensure we're finding all dependencies.
      $new_deps = $this->getConfigDependencies($known_deps);
      // Strip out any dependencies that we've already found.
      $new_deps = array_diff($new_deps, $known_deps);
      $config_deps = array_merge($config_deps, $new_deps);
    }
    // Reset the current config to our original config.
    $this->setCurrentConfig($config);
    return $config_deps;
  }

}
