<?php

namespace Drupal\config_enforce_devel\Form;

use Drupal\Component\Utility\Html;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Url;
use Drupal\config_enforce\ConfigEnforce;
use Drupal\config_enforce_devel\ConfigEnforceDevel;
use Drupal\config_enforce_devel\Form\DevelFormHelperTrait;

/**
 * Form that lists all the enforced configs, and allows them to be updated.
 */
class EnforcedConfigs extends FormBase {

  use DevelFormHelperTrait;

  const FORM_ID = 'config_enforce_enforced_configs_form';

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $this->setSharedFormProperties($form, $form_state);

    $this->addUpdateEnforcedConfigs();
    $this->addEnforcedConfigsTable();

    $this->form()['#attached']['library'][] = 'config_enforce_devel/config-enforce-devel';
    $this->form()['#attached']['library'][] = 'core/drupal.dialog.ajax';

    return $this->form();
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $this->setSharedFormProperties($form, $form_state);

    if (!$this->triggerIsUpdateSettingButton()) return;

    if (empty($this->getSelectedConfigs())) {
      $this->formState()
        ->setErrorByName('enforced_configs', $this->t('At least one enforced config must be selected.'));
    }
    if (empty($this->getUpdatedSettings())) {
      $this->formState()
        ->setErrorByName('update_enforced_configs][updated_settings', $this->t('At least one setting must be updated.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $this->setSharedFormProperties($form, $form_state);

    if (!$this->triggerIsUpdateSettingButton()) return;

    $this->configEnforce = new ConfigEnforceDevel();
    $updated_settings = $this->getUpdatedSettings();
    $enforced_configs = ConfigEnforce::getEnforcedConfigs();

    foreach ($this->getSelectedConfigs() as $config_name) {
      $new_settings = array_merge($enforced_configs[$config_name], $updated_settings);
      $this->configEnforce
        ->setConfigName($config_name)
        ->setTargetModule($new_settings['target_module'])
        ->setConfigDirectory($new_settings['config_directory'])
        ->setEnforcementLevel($new_settings['enforcement_level'])
        ->setConfigFormUri($new_settings['config_form_uri'])
        ->save();
    }
  }

  /**
   * Determine whether the form was submitted with the "Update settings" button.
   */
  protected function triggerIsUpdateSettingButton() {
    $button = $this->formState()->getTriggeringElement();
    return $button['#value'] == "Update Settings";
  }

  /**
   * {@inheritdoc}
   *
   * This provides an easy way for enforced configs to be enforced themselves.
   */
  public function getEditableConfigNames() {
    return ['config_enforce.enforced_configs'];
  }

  /**
   * Add fields to update config enforce settings for multiple enforced configs.
   */
  protected function addUpdateEnforcedConfigs() {
    $this->form()['update_enforced_configs'] = [
      '#type' => 'details',
      '#title' => 'Update settings',
      '#tree' => TRUE,
    ];
    $this->form()['update_enforced_configs']['updated_settings'] = [
      '#type' => 'container',
      '#tree' => TRUE,
    ];

    $this
      ->addUpdateTargetModule()
      ->addUpdateConfigDirectory()
      ->addUpdateEnforcementLevel();

    $this->form()['update_enforced_configs']['submit'] = [
      '#type' => 'submit',
 #     '#name' => 'update_enforced_configs',
      '#value' => 'Update settings',
      '#submit' => ['::submitForm'],
    ];
  }

  /**
   * Add field to update the target module for multiple enforced configs.
   */
  protected function addUpdateTargetModule() {
    $this->form()['update_enforced_configs']['updated_settings']['target_module'] = $this->getTargetModuleField();
    $this->form()['update_enforced_configs']['updated_settings']['target_module']['#options'] = ['no-op' => 'No change']
      + $this->form()['update_enforced_configs']['updated_settings']['target_module']['#options'];
    $this->form()['update_enforced_configs']['updated_settings']['target_module']['#default_value'] = 'no-op';
    return $this;
  }

  /**
   * Add fields to update the config directory for multiple enforced configs.
   */
  protected function addUpdateConfigDirectory() {
    $this->form()['update_enforced_configs']['updated_settings']['config_directory'] = $this->getConfigDirectoryField();
    $this->form()['update_enforced_configs']['updated_settings']['config_directory']['#options'] = ['no-op' => 'No change']
      + $this->form()['update_enforced_configs']['updated_settings']['config_directory']['#options'];
    $this->form()['update_enforced_configs']['updated_settings']['config_directory']['#default_value'] = 'no-op';
    return $this;
  }

  /**
   * Add fields to update the enforcement level for multiple enforced configs.
   */
  protected function addUpdateEnforcementLevel() {
    $this->form()['update_enforced_configs']['updated_settings']['enforcement_level'] = $this->getEnforcementLevelField();
    $this->form()['update_enforced_configs']['updated_settings']['enforcement_level']['#options'] = ['no-op' => 'No change']
      + $this->form()['update_enforced_configs']['updated_settings']['enforcement_level']['#options'];
    $this->form()['update_enforced_configs']['updated_settings']['enforcement_level']['#default_value'] = 'no-op';
    return $this;
  }

  /**
   * Add table to select multiple enforced configs to allow their settings to be updated.
   */
  protected function addEnforcedConfigsTable() {
    $this->form()['enforced_configs'] = [
      '#type' => 'tableselect',
      '#header' => $this->getHeader(),
      '#options' => $this->getOptions(),
      '#empty' => $this->t('No enforced config objects found'),
    ];
  }

  /**
   * Return a table header render array.
   */
  protected function getHeader() {
    return [
      'config_name' => ['data' => $this->t('Configuration object name')],
      'enforcement_level' => ['data' => $this->t('Enforcement level')],
      'target_module' => ['data' => $this->t('Target module')],
      'config_directory' => ['data' => $this->t('Config directory')],
      'settings' => ['data' => $this->t('Config enforce settings')],
    ];
  }

  /**
   * Return a tableselect options render array.
   */
  protected function getOptions() {
    $rows = [];
    foreach ($this->getEnforcedConfigs() as $config_name) {
      $this->setCurrentConfig($config_name);

      $settings = [
        'data' => [
          '#type' => 'link',
          '#title' => $this->t('Settings'),
          '#url' => Url::fromRoute('config_enforce_devel.config_edit', ['arg' => $config_name]),
          '#attributes' => [
            'class' => [
              'use-ajax',
              'button',
            ],
          ],
        ],
      ];

      $rows[$config_name] = [
        '#attributes' => [
          'class' => $this->getRowClasses(),
        ],
        'config_name' => $this->getConfigNameLink(),
        'enforcement_level' => ConfigEnforce::getEnforcementLevelLabel($config_name),
        'target_module' => ConfigEnforce::getTargetModule($config_name),
        'config_directory' => ConfigEnforce::getConfigDirectory($config_name),
        'settings' => $settings,
      ];
    }

    return $rows;
  }

  /**
   * Return the config name as a link to its form, if the URI is available.
   */
  protected function getConfigNameLink() {
    $config_name = $this->getCurrentConfig();
    $uri = ConfigEnforce::getConfigFormUri($config_name);
    if (empty($uri)) return $config_name;
    return Link::fromTextAndUrl($config_name, Url::fromUri('internal:' . $uri));
  }

  /**
   * Return row class to allow for better styling.
   */
  protected function getRowClasses() {
    $config_name = $this->getCurrentConfig();
    $classes = [];
    if (ConfigEnforce::isEnforced($config_name)) {
      $classes[] = 'enforced';
      $classes[] = 'level-' . ConfigEnforce::getEnforcementLevel($config_name);
    }
    return $classes;
  }

  /**
   * List only enforced config object names.
   */
  protected function getEnforcedConfigs() {
    $enforced = [];
    $enforced_configs = ConfigEnforce::getEnforcedConfigs();
    ksort($enforced_configs);
    foreach ($enforced_configs as $config_name => $setting) {
      $enforced[] = $config_name;
    }
    return $enforced;
  }

  /**
   * Return a list of selected enforced configs.
   */
  protected function getSelectedConfigs() {
    return array_filter($this->formState()->getValue('enforced_configs'));
  }

  /**
   * Return an array of settings to update and the updated values.
   */
  protected function getUpdatedSettings() {
    $updated_settings = $this->formState()->getValue('update_enforced_configs')['updated_settings'];
    foreach ($updated_settings as $setting => $value) {
      if ($value == 'no-op') {
        unset($updated_settings[$setting]);
      }
    }
    return $updated_settings;
  }

}
